import json
import os
from pathlib import Path, PurePath

from string import Template

from .configuration import PARENT_KEY, ConstrictorConfiguration

BASE_CONFIG_ENV_KEY = "CONSTRICTOR_BUILD_BASE_CONFIG_PATH"
BASE_CONFIG_PATH_DEFAULT = "~/constrictor-build-config.json"


def is_str(s):
    return isinstance(s, str)


def resolve_base_config_path(base_config_path):
    if base_config_path:
        return base_config_path

    if BASE_CONFIG_ENV_KEY in os.environ:
        return os.environ[BASE_CONFIG_ENV_KEY]

    return BASE_CONFIG_PATH_DEFAULT


class ConfigReader(object):
    def __init__(self, config_path, base_config_path=None):
        self.config_path = config_path
        self.base_config_path = Path(str(Path(resolve_base_config_path(base_config_path)).expanduser())).absolute()

    @staticmethod
    def read_json(path):
        with open(path) as config_fp:
            return json.load(config_fp)

    def load_base_config(self):
        if str(self.base_config_path) and Path(self.base_config_path).exists():
            return self.read_json(self.base_config_path)

        return {}

    def load_config_file(self, config_path):
        """
        Load a tuple of the (parent_config_path[str], config_data [dict]). parent_config_path will be None to indicate
        this is the root config (however the base_config_path config will always be the root if set).
        """
        config = self.read_json(config_path)
        return config.get(PARENT_KEY), config

    def get_configuration(self):
        resolved_configuration = ConstrictorConfiguration(self.load_base_config())

        configuration_list = []
        seen_paths = set()

        current_config_paths = [self.config_path]

        while len(current_config_paths):
            current_config_path = os.path.realpath(current_config_paths.pop(0))
            if current_config_path in seen_paths:
                raise RuntimeError(
                    "{} already loaded, possible recursive config parents.".format(current_config_paths))
            relative_parent_config_paths, configuration = self.load_config_file(current_config_path)

            configuration_list.insert(0, configuration)

            seen_paths.add(current_config_path)

            if relative_parent_config_paths is not None:
                if not isinstance(relative_parent_config_paths, list):
                    relative_parent_config_paths = [relative_parent_config_paths]

                for relative_parent_config_path in relative_parent_config_paths:
                    absolute_parent_config_path = PurePath(
                        f'{PurePath(current_config_path).parent}/{Path(relative_parent_config_path).expanduser()}')
                    current_config_paths.insert(0, str(absolute_parent_config_path))

        for configuration in configuration_list:
            resolved_configuration.update_configuration(configuration)

        resolved_configuration.interpolate_variables()
        resolved_configuration.interpolate_configuration_values()

        return resolved_configuration

    @staticmethod
    def parse_environment_variable_value(value):
        t = Template(value)
        return t.substitute(os.environ)

    def interpolate_environment_variables_in_list(self, value_list):
        for index, value in enumerate(value_list):
            if isinstance(value, dict):
                self.interpolate_environment_variables_in_dict(value)
            elif isinstance(value, list):
                self.interpolate_environment_variables_in_list(value)
            elif is_str(value):
                value_list[index] = self.parse_environment_variable_value(value)

    def interpolate_environment_variables_in_dict(self, value_dict):
        for key, value in value_dict.items():
            if isinstance(value, dict):
                self.interpolate_environment_variables_in_dict(value)
            elif isinstance(value, list):
                self.interpolate_environment_variables_in_list(value)
            elif is_str(value):
                value_dict[key] = self.parse_environment_variable_value(value)



from functools import partial
from os import environ
from string import Template

from .dpkg import LINK_PATH_KEY

PARENT_KEY = "parent"
DEB_CONSTRICTOR_KEY = "deb_constrictor"
IGNORE_PATHS_KEY = "ignore_paths"
VARIABLES_KEY = "variables"
ENVIRONMENT_VARIABLES_KEY = "environment_variables"
EXTRA_CONTROL_FIELDS_KEY = "extra_control_fields"
DEPENDS_KEY = "Depends"
PROVIDES_KEY = "Provides"
DIRECTORIES_KEY = "directories"
LINKS_KEY = "links"
COMMANDS_KEY = "commands"
MAINTAINER_SCRIPTS_KEY = "maintainer_scripts"
CONFIG_FILES_KEY = "configuration_files"
DIRECTORY_PATH_KEYS = ('source', 'destination')


def ensure_trailing_slash(path):
    """Return path if path ends with a / or path + / if it doesn't."""
    return path if path.endswith('/') else path + '/'


def ensure_trailing_slashes_in_directory(directory):
    """Return a directory entry where all path entries will be guaranteed to have a trailing /"""
    new_directory = {}
    new_directory.update(directory)
    for path_type in DIRECTORY_PATH_KEYS:
        new_directory[path_type] = ensure_trailing_slash(directory[path_type])

    return new_directory


def extract_directory_paths(directory):
    """Return a dictionary for a directory containing only the path items."""
    return {k: v for k, v in directory.items() if k in DIRECTORY_PATH_KEYS}


def directory_entries_equal(dir1, dir2):
    """Compare two directory entries, considered equal if only the path items match."""
    dir1_paths, dir2_paths = map(extract_directory_paths, map(ensure_trailing_slashes_in_directory, (dir1, dir2)))
    return dir1_paths == dir2_paths


def directory_entries_not_equal(dir1, dir2):
    """Negate directory_entries_equal function (for use when inverse is required in map()/filter() call)"""
    return not directory_entries_equal(dir1, dir2)


def interpolate_value(v, context):
    if isinstance(v, dict):
        interpolate_dictionary(v, context)
        return None
    elif isinstance(v, list):
        interpolate_list(v, context)
        return None
    elif isinstance(v, (bool, int, float)):
        return None
    else:
        return Template(v).substitute(context)


def interpolate_list(list_, context):
    """Walk through list and interpolate variables for each value."""
    for i, v in enumerate(list_):
        interpolated = interpolate_value(v, context)
        if interpolated is not None:
            list_[i] = interpolated


def interpolate_dictionary(d, context):
    """Walk through dictionary and interpolate variables for each value."""
    for k, v in d.items():
        if k == COMMANDS_KEY:  # don't interpolate commands at build time as some variables are passed in at run time
            continue
        interpolated = interpolate_value(v, context)
        if interpolated is not None:
            d[k] = interpolated


class ConstrictorConfiguration(object):
    """
    Configuration for the DpkgBuilder. Should be instantiated with the root config dict, and then updated with child
    (overriding) configuration dictionaries by calling the update_configuration method with reach child.
    """

    def __init__(self, base_configuration):
        self.configuration = {}
        self.environment_variables = {}
        self.variables = {}
        self.update_configuration(base_configuration)

    def update_configuration(self, configuration):
        """
        Override the existing configuration with the new values. Is more advanced than just doing old[k] = new[k] as it
        is also aware of what items are lists and should be appended to.
        """
        for k, v in configuration.items():
            if k == PARENT_KEY:
                continue
            elif k == DEB_CONSTRICTOR_KEY:
                self.update_deb_constrictor_configuration(v)
            elif k == EXTRA_CONTROL_FIELDS_KEY:
                self.update_extra_control_fields(v)
            elif k == DIRECTORIES_KEY:
                self.update_directory_entries(v)
            elif k == LINKS_KEY:
                self.update_link_entries(v)
            elif k == MAINTAINER_SCRIPTS_KEY:
                self.update_maintainer_scripts(v)
            elif k == CONFIG_FILES_KEY:
                self.update_configuration_files(v)
            else:
                self.configuration[k] = v

    def update_deb_constrictor_configuration(self, configuration):
        """
        Updates items in the DEB_CONSTRICTOR_KEY value dict with new items. Will append items to lists where
        appropriate.
        """
        if DEB_CONSTRICTOR_KEY not in self.configuration:
            self.configuration[DEB_CONSTRICTOR_KEY] = {}

        for k, v in configuration.items():
            if k == IGNORE_PATHS_KEY:
                self.update_ignore_paths(v)
            elif k in (VARIABLES_KEY, ENVIRONMENT_VARIABLES_KEY):
                self.update_variables(k, v)
            elif k == COMMANDS_KEY:
                self.update_commands(v)
            else:
                self.configuration[DEB_CONSTRICTOR_KEY][k] = v

    def update_ignore_paths(self, ignore_paths):
        """
        Updates the IGNORE_PATHS_KEY list in the DEB_CONSTRICTOR_KEY dict with the passed in list. Will only add the
        path if it does not exist in the list already (no duplicates).
        """
        if IGNORE_PATHS_KEY not in self.configuration[DEB_CONSTRICTOR_KEY]:
            self.configuration[DEB_CONSTRICTOR_KEY][IGNORE_PATHS_KEY] = []

        for ignore_path in ignore_paths:
            if ignore_path not in self.configuration[DEB_CONSTRICTOR_KEY][IGNORE_PATHS_KEY]:
                self.configuration[DEB_CONSTRICTOR_KEY][IGNORE_PATHS_KEY].append(ignore_path)

    def update_extra_control_fields(self, control_fields):
        """
        Updates existing EXTRA_CONTROL_FIELDS_KEY dictionary with items from the passed in dictionary. Appends
        (uniquely) to list type items instead of overriding.
        """
        if EXTRA_CONTROL_FIELDS_KEY not in self.configuration:
            self.configuration[EXTRA_CONTROL_FIELDS_KEY] = {}

        for k, v in control_fields.items():
            if k in (DEPENDS_KEY, PROVIDES_KEY):
                self.update_extra_control_field_list(k, v)
            else:
                self.configuration[EXTRA_CONTROL_FIELDS_KEY][k] = v

    def update_extra_control_field_list(self, control_field_name, new_list):
        """
        Appends items in new_list to the given key (control_field_name) in the EXTRA_CONTROL_FIELDS_KEY dictionary.
        Makes sure items are unique.
        """
        if control_field_name not in self.configuration[EXTRA_CONTROL_FIELDS_KEY]:
            self.configuration[EXTRA_CONTROL_FIELDS_KEY][control_field_name] = []

        for field_item in new_list:
            if field_item not in self.configuration[EXTRA_CONTROL_FIELDS_KEY][control_field_name]:
                self.configuration[EXTRA_CONTROL_FIELDS_KEY][control_field_name].append(field_item)

    def update_directory_entries(self, directories_list):
        """
        Append the given directory entries in directories_list to the existing directories in the configuration.
        Existing directories are removed and replaced if they have the same source and destination as an incoming entry
        (as there may be legitimate cases to source the same destination to multiple targets or multiple sources to the
        same target [if they contain different files] so if either differ to an existing entry it will be added).
        """
        if DIRECTORIES_KEY not in self.configuration:
            self.configuration[DIRECTORIES_KEY] = []

        for new_directory in directories_list:
            self.configuration[DIRECTORIES_KEY] = list(filter(partial(directory_entries_not_equal, new_directory),
                                                              self.configuration[DIRECTORIES_KEY]))
            self.configuration[DIRECTORIES_KEY].append(new_directory)

    def update_link_entries(self, links_list):
        """
        Append the given links in the links_list to the existing links in the configuration. Existing link entries are
        removed and replaced with incoming if they match on the source field only (as it doesn't make sense to create
        two links in the same place).
        """
        if LINKS_KEY not in self.configuration:
            self.configuration[LINKS_KEY] = []

        for new_link in links_list:
            self.configuration[LINKS_KEY] = list(filter(lambda link: link[LINK_PATH_KEY] != new_link[LINK_PATH_KEY],
                                                        self.configuration[LINKS_KEY]))
            self.configuration[LINKS_KEY].append(new_link)

    def update_maintainer_scripts(self, maintainer_scripts):
        """Override existing maintainer script keys with the new ones."""
        if MAINTAINER_SCRIPTS_KEY not in self.configuration:
            self.configuration[MAINTAINER_SCRIPTS_KEY] = {}

        self.configuration[MAINTAINER_SCRIPTS_KEY].update(maintainer_scripts)

    def update_configuration_files(self, configuration_files):
        """Append configuration files to existing."""
        if CONFIG_FILES_KEY not in self.configuration:
            self.configuration[CONFIG_FILES_KEY] = []

        self.configuration[CONFIG_FILES_KEY] += list(
            filter(lambda path: path not in self.configuration[CONFIG_FILES_KEY], configuration_files))

    def update_variables(self, variables_key, new_variables):
        """
        Append the list of variables to the given variables_key. Does not do de-duplication of the variable name as it
        might be good to have parent variables populate and the be able to be used in a child config.
        """
        if variables_key not in self.configuration[DEB_CONSTRICTOR_KEY]:
            self.configuration[DEB_CONSTRICTOR_KEY][variables_key] = []

        self.configuration[DEB_CONSTRICTOR_KEY][variables_key] += new_variables

    def update_commands(self, commands):
        if COMMANDS_KEY not in self.configuration[DEB_CONSTRICTOR_KEY]:
            self.configuration[DEB_CONSTRICTOR_KEY][COMMANDS_KEY] = {}

        self.configuration[DEB_CONSTRICTOR_KEY][COMMANDS_KEY].update(commands)

    def interpolate_variables(self):
        """
        Should be called before variables are used, when we have finished updating all configs down the hierarchy, to
        interpolate the variables with variables before they can be used.
        """
        self.store_variable_list(VARIABLES_KEY, self.variables)
        self.store_variable_list(ENVIRONMENT_VARIABLES_KEY, self.environment_variables)

    def store_variable_list(self, variables_list_key, variables_container):
        """
        Interpolate each variable in the variables list (a list of lists, each item is [key, value]) and store it in the
        dictionary for use later, i.e: [[k, v]] => {k: interpolate_value(v)}

        Because the items are processed in order, items further along the list might be interpolated with variables set
        by earlier elements.
        """
        for k, v in self.configuration[DEB_CONSTRICTOR_KEY].get(variables_list_key, []):
            variables_container[k] = self.interpolate_value(v)

    def get_template_context(self):
        """
        Template context (for interpolating variables) is os.environ, which is overridden by self.environment_variables
        and then overridden by self.variables.

        If you wanted to be more performant you would cache the ctx and invalidate it when the var variables change, but
        I don't foresee this being an issue.
        """
        ctx = dict(environ)
        ctx.update(self.environment_variables)
        ctx.update(self.variables)
        return ctx

    def interpolate_value(self, value):
        return interpolate_value(value, self.get_template_context())

    def interpolate_configuration_values(self):
        """
        Recurse through the configuration and interpolate all the values with the template context. This should be
        called after all the configurations have been loaded (parent hierarchy resolved and updated) and then
        interpolate_variables called.
        """
        interpolate_dictionary(self.configuration, self.get_template_context())

    def __getitem__(self, item):
        return self.configuration[item]

    def get(self, item, default=None):
        return self.configuration.get(item, default)



import os
import fnmatch
import sys
from functools import partial
from io import BytesIO
import tarfile
import tempfile
import time
from pathlib import Path, PurePath

from .ar import ARWriter
from .control import FIELD_INSTALLED_SIZE
from .helpers import md5_for_path

MAINTAINER_SCRIPT_NAMES = ('preinst', 'postinst', 'prerm', 'postrm')
TAR_INFO_KEYS = ('uname', 'gname', 'uid', 'gid', 'mode')
FORCE_DIRECTORY_INCLUSION_FILENAMES = (".debkeep", ".gitkeep")
DEBIAN_BINARY_VERSION = '2.0'
TAR_DEFAULT_MODE = 0o755
AR_DEFAULT_MODE = 0o644

LINK_PATH_KEY = "path"
LINK_TARGET_KEY = "target"


def generate_directories(path, existing_dirs=None):
    """Recursively build a list of directories inside a path."""
    existing_dirs = existing_dirs or []
    directory_name = os.path.dirname(path)

    if directory_name == '.':
        return

    existing_dirs.append(directory_name)
    generate_directories(directory_name, existing_dirs)

    return existing_dirs


class DPKGBuilder(object):
    """
    Finds files to use, builds tar archive and then archives into ar format. Builds + includes debian control files.
    """

    def __init__(self, output_directory, control, data_dirs, links, maintainer_scripts=None, output_name=None,
                 ignore_paths=None, configuration_files=None):
        self.output_directory = Path(output_directory).expanduser()
        self.data_dirs = data_dirs or []
        self.links = links or {}
        self.maintainer_scripts = maintainer_scripts
        self.seen_data_dirs = set()
        self.control = control
        self.working_dir = None
        self.output_name = output_name or control.get_default_output_name()
        self.ignore_paths = ignore_paths or []
        self.configuration_files = configuration_files or []
        self.actual_config_files = []

    @staticmethod
    def path_matches_glob_list(glob_list, path):
        path_matcher = partial(fnmatch.fnmatch, path)
        return any(map(path_matcher, glob_list))

    def should_skip_path(self, path):
        return self.path_matches_glob_list(self.ignore_paths, path)

    def path_is_config(self, path):
        return self.path_matches_glob_list(self.configuration_files, path)

    def list_data_dir(self, source_dir):
        """
        Iterator to recursively list all files in a directory that should be included. Returns a tuple of absolute
        file_path (on local) and the relative path (relative to source).
        If the file name matches a filename to skip (should_skip_file returns true) it will not be returned.
        """
        for root_dir, dirs, files in os.walk(source_dir):
            for file_name in files:
                file_path = os.path.join(root_dir, file_name)
                relative_path = file_path[len(source_dir):]

                if self.should_skip_path(relative_path):
                    continue

                yield file_path, relative_path

    def add_directory_root_to_archive(self, archive, dir_conf, file_path):
        for directory in reversed(generate_directories(file_path)):
            if directory in self.seen_data_dirs:
                continue

            dir_ti = tarfile.TarInfo()
            dir_ti.type = tarfile.DIRTYPE
            dir_ti.name = directory
            dir_ti.mtime = int(time.time())
            dir_ti.mode = TAR_DEFAULT_MODE
            dir_ti = self.filter_tar_info(dir_ti, dir_conf)
            archive.addfile(dir_ti)

            self.seen_data_dirs.add(directory)

    def filter_tar_info(self, tar_info, dir_conf):
        app_name = self.control.get_control_line_value('Name')
        executable_path = '.' + str(PurePath(dir_conf['destination']
                                             ).joinpath(f'{app_name}.app').joinpath(dir_conf['executable']))
        executable_path = executable_path.replace('\\', '/')

        if tar_info.name == executable_path:
            setattr(tar_info, TAR_INFO_KEYS[4], TAR_DEFAULT_MODE)

        for tar_info_key in TAR_INFO_KEYS:
            if dir_conf.get(tar_info_key) is not None:
                setattr(tar_info, tar_info_key, dir_conf[tar_info_key])

        if 'uid' in dir_conf and 'uname' not in dir_conf:
            tar_info.uname = ''

        if 'gid' in dir_conf and 'gname' not in dir_conf:
            tar_info.gname = ''

        return tar_info

    @property
    def data_archive_path(self):
        return Path(self.working_dir).joinpath('data.tar.gz')

    @staticmethod
    def open_tar_file(path):
        tf = tarfile.open(path, 'w:gz')
        tf.format = tarfile.USTAR_FORMAT
        return tf

    def build_data_archive(self):
        data_tar_file = self.open_tar_file(self.data_archive_path)
        file_size_bytes = 0

        file_md5s = []

        for dir_conf in self.data_dirs:
            source_dir = str(Path(dir_conf['source']).expanduser())

            for source_file_path, source_file_name in self.list_data_dir(source_dir):
                archive_path = '.' + dir_conf['destination'] + source_file_name
                archive_path = archive_path.replace('\\', '/')
                if Path(source_file_path).is_symlink() and not Path(source_file_path).exists():
                    # this is a link to a file that doesn't exist but should when we deploy if we are on the same OS
                    # (e.g. venv build with docker and .deb assembled on host) so add it as a link
                    self.links.append({
                        LINK_PATH_KEY: archive_path,
                        LINK_TARGET_KEY: str(Path(source_file_path).readlink())
                    })
                else:
                    self.add_directory_root_to_archive(data_tar_file, dir_conf, archive_path)

                    if PurePath(source_file_name).name in FORCE_DIRECTORY_INCLUSION_FILENAMES:
                        continue

                    file_size_bytes += Path(source_file_path).stat().st_size

                    file_md5s.append((md5_for_path(source_file_path), archive_path))
                    data_tar_file.add(source_file_path, arcname=archive_path, recursive=False,
                                      filter=lambda ti: self.filter_tar_info(ti, dir_conf))

                    config_path = archive_path[1:] if archive_path.startswith(".") else archive_path

                    if config_path not in self.actual_config_files and self.path_is_config(config_path):
                        self.actual_config_files.append(config_path)

        for symlink_conf in self.links:
            target = symlink_conf[LINK_TARGET_KEY]
            path = symlink_conf[LINK_PATH_KEY]

            if not path.startswith('.'):
                path = '.' + path

            self.add_directory_root_to_archive(data_tar_file, symlink_conf, path)
            link_ti = self.build_link_tarinfo(symlink_conf, target, path)

            data_tar_file.addfile(link_ti)
            file_size_bytes += len(path)

        data_tar_file.close()

        return file_size_bytes, file_md5s

    def build_link_tarinfo(self, symlink_conf, target, path):
        link_ti = tarfile.TarInfo()
        link_ti.type = tarfile.SYMTYPE
        link_ti.linkname = target
        link_ti.name = path
        link_ti.mtime = int(time.time())
        link_ti = self.filter_tar_info(link_ti, symlink_conf)
        return link_ti

    @staticmethod
    def build_member_from_string(name, content):
        content_file = BytesIO(content)
        member = tarfile.TarInfo()
        member.type = tarfile.REGTYPE
        member.name = name
        member.mtime = int(time.time())
        member.uname = 'root'
        member.gname = 'root'
        member.size = len(content)
        return member, content_file

    @staticmethod
    def validate_maintainer_scripts(maintainer_scripts):
        for script_name in maintainer_scripts.keys():
            if script_name not in MAINTAINER_SCRIPT_NAMES:
                raise ValueError("Unknown maintainer script {}".format(script_name))

    @staticmethod
    def filter_maintainer_script_tar_info(tar_info):
        tar_info.uid = 0
        tar_info.gid = 0
        tar_info.mode = TAR_DEFAULT_MODE
        return tar_info

    @staticmethod
    def filter_executable_tar_info(tar_info):
        tar_info.mode = TAR_DEFAULT_MODE
        return tar_info

    @property
    def control_archive_path(self):
        return Path(self.working_dir).joinpath('control.tar.gz')

    def build_control_archive(self, control_text, file_md5s, maintainer_scripts):
        maintainer_scripts = maintainer_scripts or {}
        self.validate_maintainer_scripts(maintainer_scripts)

        control_tar = self.open_tar_file(self.control_archive_path)

        for script_name, script_path in maintainer_scripts.items():
            control_tar.add(script_path, arcname='./' + script_name, filter=self.filter_maintainer_script_tar_info)

        control_tar.addfile(*self.build_member_from_string('./control', control_text.encode()))

        md5sum_text = '\n'.join(['  '.join(md5_file_pair) for md5_file_pair in file_md5s]) + '\n'
        control_tar.addfile(*self.build_member_from_string('./md5sums', md5sum_text.encode()))

        if self.actual_config_files:
            conf_file_text = '\n'.join(self.actual_config_files) + "\n"
            control_tar.addfile(*self.build_member_from_string('./conffiles', conf_file_text.encode()))

        control_tar.close()

    def assemble_deb_archive(self, control_archive_path, data_archive_path):
        if not Path(self.output_directory).exists():
            Path(self.output_directory).mkdir()

        pkg_path = Path(self.output_directory).joinpath(self.output_name)

        with open(pkg_path, 'wb') as ar_fp:
            ar_writer = ARWriter(ar_fp)

            ar_writer.archive_text("debian-binary", "{}\n".format(DEBIAN_BINARY_VERSION), int(time.time()), 0, 0,
                                   AR_DEFAULT_MODE)
            ar_writer.archive_file(control_archive_path, int(time.time()), 0, 0, AR_DEFAULT_MODE)
            ar_writer.archive_file(data_archive_path, int(time.time()), 0, 0, AR_DEFAULT_MODE)

        return pkg_path

    def build_package(self):
        with tempfile.TemporaryDirectory() as tmpfolder:
            self.working_dir = tmpfolder
            file_size_bytes, file_md5s = self.build_data_archive()

            if not self.control.is_field_defined(FIELD_INSTALLED_SIZE):
                self.control.installed_size_bytes = file_size_bytes

            self.build_control_archive(self.control.get_control_text(), file_md5s, self.maintainer_scripts)

            return self.assemble_deb_archive(self.control_archive_path, self.data_archive_path)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
#	tcprelay.py - TCP connection relay for usbmuxd
#
# Original work Copyright (C) 2009  Hector Martin "marcan" <hector@marcansoft.com>
# Modified work Copyright (C) 2019  phx <https://github.com/phx>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 or version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
import os
import socket
import struct
import plistlib
import socketserver
from select import select

from permasigner.ps_logger import Logger, Colors


class MuxError(Exception):
    pass


class MuxVersionError(MuxError):
    pass


class SafeStreamSocket:
    def __init__(self, address, family):
        self.sock = socket.socket(family, socket.SOCK_STREAM)
        self.sock.connect(address)

    def send(self, msg):
        totalsent = 0
        while totalsent < len(msg):
            sent = self.sock.send(bytes(msg)[totalsent:])
            if sent == 0:
                raise MuxError("socket connection broken")
            totalsent = totalsent + sent

    def recv(self, size):
        msg = b''
        while len(msg) < size:
            chunk = self.sock.recv(size - len(msg))
            if chunk == b'':
                raise MuxError("socket connection broken")
            msg = msg + chunk
        return msg


class MuxDevice(object):
    def __init__(self, devid, usbprod, serial, location):
        self.devid = devid
        self.usbprod = usbprod
        self.serial = serial
        self.location = location

    def __str__(self):
        return f"<MuxDevice: ID {self.devid} ProdID {format(self.usbprod, '#04x')} Serial '{self.serial}' " \
               f"Location {format(self.location, '#x')}"


class BinaryProtocol(object):
    TYPE_RESULT = 1
    TYPE_CONNECT = 2
    TYPE_LISTEN = 3
    TYPE_DEVICE_ADD = 4
    TYPE_DEVICE_REMOVE = 5
    VERSION = 0

    def __init__(self, socket):
        self.socket = socket
        self.connected = False

    def _pack(self, req, payload):
        if req == self.TYPE_CONNECT:
            return struct.pack("IH", payload['DeviceID'], payload['PortNumber']) + b"\x00\x00"
        elif req == self.TYPE_LISTEN:
            return ""
        else:
            raise ValueError(f"Invalid outgoing request type {req}")

    def _unpack(self, resp, payload):
        if resp == self.TYPE_RESULT:
            return {'Number': struct.unpack("I", payload)[0]}
        elif resp == self.TYPE_DEVICE_ADD:
            devid, usbpid, serial, pad, location = struct.unpack(
                "IH256sHI", payload)
            serial = serial.decode().split("\0")[0]
            return {'DeviceID': devid,
                    'Properties': {'LocationID': location, 'SerialNumber': serial, 'ProductID': usbpid}}
        elif resp == self.TYPE_DEVICE_REMOVE:
            devid = struct.unpack("I", payload)[0]
            return {'DeviceID': devid}
        else:
            raise MuxError(f"Invalid incoming request type {resp}")

    def sendpacket(self, req, tag, payload=None):
        if payload is None:
            payload = {}
        payload = self._pack(req, payload)
        if not isinstance(payload, bytes):
            payload = bytes(payload, 'utf-8')
        if self.connected:
            raise MuxError("Mux is connected, cannot issue control packets")
        length = 16 + len(payload)
        data = struct.pack('4I', length, self.VERSION, req, tag) + payload
        self.socket.send(data)

    def getpacket(self):
        if self.connected:
            raise MuxError("Mux is connected, cannot issue control packets")
        dlen = self.socket.recv(4)
        dlen = struct.unpack("I", dlen)[0]
        body = self.socket.recv(dlen - 4)
        version, resp, tag = struct.unpack("3I", body[:0xc])
        if version != self.VERSION:
            raise MuxVersionError(
                f"Version mismatch: expected {self.VERSION}, got {version}")
        payload = self._unpack(resp, body[0xc:])
        return resp, tag, payload


class PlistProtocol(BinaryProtocol):
    TYPE_RESULT = "Result"
    TYPE_CONNECT = "Connect"
    TYPE_LISTEN = "Listen"
    TYPE_DEVICE_ADD = "Attached"
    TYPE_DEVICE_REMOVE = "Detached"
    TYPE_PLIST = 8
    VERSION = 1

    def __init__(self, socket):
        BinaryProtocol.__init__(self, socket)

    def _pack(self, req, payload):
        return payload

    def _unpack(self, resp, payload):
        return payload

    def sendpacket(self, req, tag, payload=None):
        if payload is None:
            payload = {}
        payload['ClientVersionString'] = 'usbmux.py by marcan - python2/3 compatibility by phx'
        if isinstance(req, int):
            req = [self.TYPE_CONNECT, self.TYPE_LISTEN][req - 2]
        payload['MessageType'] = req
        payload['ProgName'] = 'tcprelay'
        BinaryProtocol.sendpacket(
            self, self.TYPE_PLIST, tag, plistlib.dumps(payload))

    def getpacket(self):
        resp, tag, payload = BinaryProtocol.getpacket(self)
        if resp != self.TYPE_PLIST:
            raise MuxError(f"Received non-plist type {resp}")
        payload = plistlib.loads(payload)
        return payload['MessageType'], tag, payload


class MuxConnection(object):
    def __init__(self, socketpath, protoclass):
        self.socketpath = socketpath
        if os.environ.get("HOST_IS_WINDOWS", False):
            family = socket.AF_INET
            address = ('host.docker.internal', 27015)
        elif os.name == 'nt':
            family = socket.AF_INET
            address = ('127.0.0.1', 27015)
        else:
            family = socket.AF_UNIX
            address = self.socketpath
        self.socket = SafeStreamSocket(address, family)
        self.proto = protoclass(self.socket)
        self.pkttag = 1
        self.devices = []

    def _getreply(self):
        while True:
            resp, tag, data = self.proto.getpacket()
            if resp == self.proto.TYPE_RESULT:
                return tag, data
            else:
                raise MuxError(f"Invalid packet type received: {resp}")

    def _processpacket(self):
        resp, tag, data = self.proto.getpacket()
        if resp == self.proto.TYPE_DEVICE_ADD:
            self.devices.append(
                MuxDevice(data['DeviceID'], data['Properties']['ProductID'], data['Properties']['SerialNumber'],
                          data['Properties']['LocationID']))
        elif resp == self.proto.TYPE_DEVICE_REMOVE:
            for dev in self.devices:
                if dev.devid == data['DeviceID']:
                    self.devices.remove(dev)
        elif resp == self.proto.TYPE_RESULT:
            raise MuxError(f"Unexpected result: {resp}")
        else:
            raise MuxError(f"Invalid packet type received: {resp}")

    def _exchange(self, req, payload=None):
        if payload is None:
            payload = {}
        mytag = self.pkttag
        self.pkttag += 1
        self.proto.sendpacket(req, mytag, payload)
        recvtag, data = self._getreply()
        if recvtag != mytag:
            raise MuxError('Replay tag mismatch: expected',
                           str(mytag), 'received', str(recvtag))
        return data['Number']

    def listen(self):
        ret = self._exchange(self.proto.TYPE_LISTEN)
        if ret != 0:
            raise MuxError(f"Listen failed: error {ret}")

    def process(self, timeout=None):
        if self.proto.connected:
            raise MuxError(
                "Socket is connected, cannot process listener events")
        rlo, wlo, xlo = select([self.socket.sock], [], [
            self.socket.sock], timeout)
        if xlo:
            self.socket.sock.close()
            raise MuxError("Exception in listener socket")
        if rlo:
            self._processpacket()

    def connect(self, device, port):
        ret = self._exchange(self.proto.TYPE_CONNECT,
                             {'DeviceID': device.devid, 'PortNumber': ((port << 8) & 0xFF00) | (port >> 8)})
        print(((port << 8) & 0xFF00) | (port >> 8))
        if ret != 0:
            raise MuxError(f"Connect failed: error {ret}")
        self.proto.connected = True
        return self.socket.sock

    def close(self):
        self.socket.sock.close()


class USBMux(object):
    def __init__(self, socketpath):
        self.socketpath = socketpath
        self.listener = MuxConnection(self.socketpath, BinaryProtocol)
        try:
            self.listener.listen()
            self.version = 0
            self.protoclass = BinaryProtocol
        except MuxVersionError:
            self.listener = MuxConnection(self.socketpath, PlistProtocol)
            self.listener.listen()
            self.protoclass = PlistProtocol
            self.version = 1
        self.devices = self.listener.devices

    def process(self, timeout=None):
        self.listener.process(timeout)

    def connect(self, device, port):
        connector = MuxConnection(self.socketpath, self.protoclass)
        return connector.connect(device, port)


class SocketRelay(object):
    def __init__(self, a, b, maxbuf=65535):
        self.a = a
        self.b = b
        self.atob = b''
        self.btoa = b''
        self.maxbuf = maxbuf

    def handle(self):
        while True:
            try:
                rlist = []
                wlist = []
                xlist = [self.a, self.b]
                if self.atob:
                    wlist.append(self.b)
                if self.btoa:
                    wlist.append(self.a)
                if len(self.atob) < self.maxbuf:
                    rlist.append(self.a)
                if len(self.btoa) < self.maxbuf:
                    rlist.append(self.b)
                rlo, wlo, xlo = select(rlist, wlist, xlist)
                if xlo:
                    return
                if self.a in wlo:
                    n = self.a.send(self.btoa)
                    self.btoa = self.btoa[n:]
                if self.b in wlo:
                    n = self.b.send(self.atob)
                    self.atob = self.atob[n:]
                if self.a in rlo:
                    s = self.a.recv(self.maxbuf - len(self.atob))
                    if not s:
                        return
                    self.atob += s
                if self.b in rlo:
                    s = self.b.recv(self.maxbuf - len(self.btoa))
                    if not s:
                        return
                    self.btoa += s
            except ConnectionResetError:
                return


class TCPRelay(socketserver.BaseRequestHandler):
    def handle(self):
        logger = Logger(self.server.args)
        logger.debug(f"Incoming connection to {self.server.server_address[1]}")
        mux = USBMux(self.server.socketpath)
        logger.debug("Waiting for devices...")
        if not mux.devices:
            mux.process(1.0)
        if not mux.devices:
            logger.debug("No device found")
            self.request.close()
            return
        dev = mux.devices[0]
        logger.debug(f"Connecting to device {str(dev)}")
        dsock = mux.connect(dev, self.server.rport)
        lsock = self.request
        logger.debug("Connection established, relaying data")
        try:
            fwd = SocketRelay(dsock, lsock, 131072)
            fwd.handle()
        finally:
            dsock.close()
            lsock.close()
        logger.debug("Connection closed")


class TCPServer(socketserver.TCPServer):
    allow_reuse_address = True


class Relayer(object):
    def __init__(self, rport, lport, args, host=None, socketpath=None):
        self.rport = rport
        self.lport = lport
        self.args = args
        if host is None:
            host = 'localhost'
        self.host = host
        if socketpath is None:
            socketpath = '/var/run/usbmuxd'
        self.socketpath = socketpath
        self.logger = Logger(self.args)

    def relay(self):
        self.logger.log(f"Creating listening port {self.lport} for device port {self.rport}", color=Colors.yellow)
        server = TCPServer((self.host, self.lport), TCPRelay)
        server.rport = self.rport
        server.socketpath = self.socketpath
        server.args = self.args
        servers = [server]

        alive = True

        while alive:
            try:
                rl, wl, xl = select(servers, [], [])
                for serv in rl:
                    serv.handle_request()
            except:
                alive = False



from pathlib import PurePath, Path
from shutil import copyfileobj

AR_HEADER_TEXT = "!<arch>\n"


class ARWriter(object):
    def __init__(self, fp):
        self.archive_ready = False
        self.fp = fp

    def write_header(self):
        if not self.archive_ready:
            raise RuntimeError("Archive not yet created.")
        self.fp.write(AR_HEADER_TEXT.encode("ascii"))

    def create_archive(self):
        if self.archive_ready:
            raise RuntimeError("Archive has already been created.")
        self.archive_ready = True
        self.write_header()

    def _write_file_header(self, file_name, mod_timestamp, uid, gid, mode, file_size):
        if not self.archive_ready:
            self.create_archive()

        mode_string = '{:04o}'.format(mode)

        file_header = '{: <16}{: <12}{: <6}{: <6}{: <8}{: <10}`\n'.format(file_name, mod_timestamp, uid, gid,
                                                                          mode_string, file_size)

        self.fp.write(file_header.encode('ascii'))

    def _write_file_content(self, source_path):
        with open(source_path, 'rb') as source_file:
            copyfileobj(source_file, self.fp)

    def _write_file_alignment(self):
        """ar files are aligned to 2n bytes, this will write an extra byte to make sure they line up."""
        self.fp.write("\n".encode("ascii"))

    def _write_text(self, text):
        self.fp.write(text.encode("ascii"))

    def archive_file(self, file_path, mod_time_override=None, uid_override=None, gid_override=None, mode_override=None):
        if not self.archive_ready:
            self.create_archive()
        file_to_archive_name = PurePath(file_path).name
        file_stat = Path(file_path).stat()

        mod_time = mod_time_override or file_stat.st_mtime
        uid = uid_override or file_stat.st_uid
        gid = gid_override or file_stat.st_gid
        mode = mode_override or file_stat.st_mode
        file_size = file_stat.st_size
        self._write_file_header(file_to_archive_name, mod_time, uid, gid, mode, file_size)
        self._write_file_content(file_path)

        if file_size % 2 != 0:
            self._write_file_alignment()

    def archive_text(self, file_name, text, mod_time, uid, gid, mode):
        if not self.archive_ready:
            self.create_archive()
        file_size = len(text)
        self._write_file_header(file_name, mod_time, uid, gid, mode, file_size)
        self._write_text(text)

        if file_size % 2 != 0:
            self._write_file_alignment()



import pkgutil
from urllib.parse import urlparse

WINDOWS_LINE_ENDING = b'\r\n'
UNIX_LINE_ENDING = b'\n'


class Copier:
    def __init__(self, app_name, app_bundle, app_version, app_min_ios, app_author, in_package):
        self.app_name = app_name
        self.app_bundle = app_bundle
        self.app_version = app_version
        self.app_min_ios = app_min_ios
        self.app_author = app_author
        self.in_package = in_package

    def copy_postinst(self, file_path):
        """Copy postinst file.

        Args:
            file_path (String): Path of the copy destination.
        """

        # Read the file
        if self.in_package:
            filedata = pkgutil.get_data(__name__, "data/DEBIAN/postinst")
        else:
            with open("permasigner/data/DEBIAN/postinst", 'rb') as file:
                filedata = file.read()

        # Replace the target string
        filedata = filedata.replace(b"{APP_NAME}", self.app_name.encode())
        filedata = filedata.replace(WINDOWS_LINE_ENDING, UNIX_LINE_ENDING)

        # Write the file out again
        with open(file_path, 'wb') as file:
            for line in filedata.splitlines():
                file.write(line + b'\n')

    def copy_postrm(self, file_path):
        """Copy postrm file.

        Args:
            file_path (String): Path of the copy destination.
        """

        # Read the file
        if self.in_package:
            filedata = pkgutil.get_data(__name__, "data/DEBIAN/postrm")
        else:
            with open('permasigner/data/DEBIAN/postrm', 'rb') as file:
                filedata = file.read()

        # Replace the target string
        filedata = filedata.replace(b"{APP_NAME}", self.app_name.encode())
        filedata = filedata.replace(WINDOWS_LINE_ENDING, UNIX_LINE_ENDING)

        # Write the file out again
        with open(file_path, 'wb') as file:
            file.write(filedata)

    def copy_control(self, file_path):
        """Copy control file.

        Args:
            file_path (String): Path of the copy destination.
        """

        # Read the file
        if self.in_package:
            filedata = pkgutil.get_data(__name__, "data/DEBIAN/control")
        else:
            with open('permasigner/data/DEBIAN/control', 'rb') as file:
                filedata = file.read()

        # Replace the target strings
        filedata = filedata.replace(b"{APP_NAME}", self.app_name.encode())
        filedata = filedata.replace(b"{APP_NAME_ENCODED}", urlparse(self.app_name).path.encode())
        filedata = filedata.replace(b"{APP_BUNDLE}", self.app_bundle.encode())
        filedata = filedata.replace(b"{APP_VERSION}", self.app_version.encode())
        filedata = filedata.replace(b"{APP_MIN_IOS}", self.app_min_ios.encode())
        filedata = filedata.replace(b"{APP_AUTHOR}", self.app_author.encode())
        filedata = filedata.replace(WINDOWS_LINE_ENDING, UNIX_LINE_ENDING)

        # Write the file out again
        with open(file_path, 'wb') as file:
            file.write(filedata)

    def copy_entitlements(self, file_path):
        """Copy entitlements file.

        Args:
            file_path (String): Path of the copy destination.
        """

        # Read the file
        if self.in_package:
            filedata = pkgutil.get_data(__name__, "data/entitlements.plist").decode('utf_8')
        else:
            with open('permasigner/data/entitlements.plist', 'r') as file:
                filedata = file.read()

        # Replace the target string
        if "science.xnu.undecimus" in self.app_bundle:
            filedata = filedata.replace("{APP_BUNDLE}", "8H69ZMY835.science.xnu.undecimus.8H69ZMY835")
            filedata = filedata.replace("{APP_KEYCHAIN_GROUPS}", "8H69ZMY835.*")
        else:
            filedata = filedata.replace("{APP_BUNDLE}", self.app_bundle)
            filedata = filedata.replace("{APP_KEYCHAIN_GROUPS}", self.app_bundle)

        # Write the file out again
        with open(file_path, 'w') as file:
            file.write(filedata)



import tarfile
import unix_ar
from urllib.parse import urlparse
from .ps_utils import Utils
from .constrictor.control import BinaryControl
from .constrictor.dpkg import DPKGBuilder


class Control:
    def __init__(self, package, version, min_ios, name, author, executable):
        self.package = package
        self.version = version
        self.section = "Applications"
        self.arch = "iphoneos-arm"
        self.min_ios = min_ios
        self.depends = f"firmware (>={self.min_ios})"
        self.name = name
        self.description = f"{self.name} resigned with Linus Henze's CoreTrust bypass so it doesn't expire, and will persist in stock."
        self.author = author
        self.maintainer = self.author
        self.tags = f"compatible_min::ios{self.min_ios}"
        self.name_encoded = urlparse(self.name).path
        self.depiction = f"https://permasigner-depictions.itsnebula.net/depiction?name={urlparse(self.name).path}"
        self.executable = executable


class Deb(object):
    def __init__(self, source, output, args):
        self.source = source
        self.output = output
        self.args = args
        self.utils = Utils(self.args)

    def build(self, postinst, postrm, control):
        dirs = [
            {
                'source': self.source,
                'destination': '/Applications',
                'executable': control.executable

            }
        ]

        scripts = {
            'postinst': postinst,
            'postrm': postrm
        }

        links = []

        c = BinaryControl(control.package, control.version, control.arch, control.maintainer, control.description)
        c.set_control_fields({'Name': control.name,
                              'Author': control.author,
                              'Section': control.section,
                              'Depends': control.depends,
                              'Tags': control.tags,
                              'Depiction': control.depiction
                              })
        output_name = control.name + '_' + control.version + '.deb'
        d = DPKGBuilder(self.output, c, dirs, links, scripts, output_name=output_name)
        d.build_package()
        return d.output_name

    def extract(self):
        ar_file = unix_ar.open(str(self.source[0]))
        try:
            try:
                tarball = ar_file.open('data.tar.xz')
            except KeyError:
                tarball = ar_file.open('data.tar.gz')
        except KeyError:
            tarball = ar_file.open('data.tar.bz2')
        with tarfile.open(fileobj=tarball) as data:
            data.extractall(path=self.output)
        ar_file.close()



import time
from getpass import getpass
from threading import Thread

from paramiko.client import AutoAddPolicy, SSHClient
from paramiko.ssh_exception import AuthenticationException, SSHException, NoValidConnectionsError
from scp import SCPClient

from .ps_logger import Logger, Colors
from .ps_tcprelay import Relayer


class Installer:
    def __init__(self, args, path):
        self.args = args
        self.path = path
        self.logger = Logger(self.args)

    def install_deb(self):
        if self.args.tcprelay:
            split = self.args.tcprelay.split(':')
            values = dict(enumerate(split))
            rport = int(values.get(0))
            lport = int(values.get(1))
            host = values.get(2)
            socketpath = values.get(3)
        else:
            rport = 22
            lport = 2222
            host = 'localhost'
            socketpath = '/var/run/usbmuxd'

        relayer = Relayer(rport, lport, self.args, host, socketpath)
        thread = Thread(target=relayer.relay, daemon=True)
        thread.start()
        time.sleep(1)

        password = getpass(
            prompt="Please provide your user password (default = alpine): ")
        if len(password.strip()) == 0:
            password = 'alpine'

        with SSHClient() as client:
            client.set_missing_host_key_policy(AutoAddPolicy())
            try:
                client.connect(host or 'localhost',
                               port=lport,
                               username='mobile',
                               password=f'{password}',
                               timeout=5000,
                               allow_agent=False,
                               look_for_keys=False,
                               compress=True)
            except (SSHException, NoValidConnectionsError, AuthenticationException) as e:
                self.logger.error(e)
                return False

            with SCPClient(client.get_transport()) as scp:
                self.logger.log(f"Sending {self.path} to device", color=Colors.yellow)
                filename = self.path.name
                self.logger.debug(f"Copying via scp from {self.path} to /var/mobile/Documents/")
                scp.put(f'{self.path}',
                        remote_path='/var/mobile/Documents')

            self.logger.debug(f"Running command: sudo -nv")
            stdin, stdout, stderr = client.exec_command('sudo -nv')
            status = stdout.channel.recv_exit_status()
            out = stderr.read()

            if "password".encode() in out:
                command = f"sudo dpkg -i /var/mobile/Documents/{filename}"
                self.logger.debug(f"Running command: {command}")
                stdin, stdout, stderr = client.exec_command(
                    f"{command}", get_pty=True)
                time.sleep(0.2)
                stdin.write(f'{password}\n')
                stdin.flush()
                self.logger.log("Installing... this may take some time", color=Colors.yellow)

                self.logger.debug(stdout.read().decode())
                self.logger.debug(f"Running command: sudo apt-get install -f")

                streams = client.exec_command(
                    'sudo apt-get install -f', get_pty=True)
                time.sleep(0.2)
                streams[0].write(f'{password}\n')
                streams[0].flush()
                self.logger.debug(streams[1].read().decode())
            elif status == 0:
                command = f"sudo dpkg -i /var/mobile/Documents/{filename}"
                self.logger.debug(f"Running command: {command}")
                output = client.exec_command(f'{command}')[1]
                self.logger.log("Installing... this may take some time", color=Colors.yellow)
                self.logger.debug(output.read().decode())
                self.logger.debug(f"Running command: sudo apt-get install -f")

                output = client.exec_command(
                    'sudo apt-get install -f')[1]
                self.logger.debug(output.read().decode())
            else:
                command = f"su root -c 'dpkg -i /var/mobile/Documents/{filename}'"
                self.logger.debug(f"Running command: {command}")

                streams = client.exec_command(
                    f"{command}", get_pty=True)

                output = streams[1].channel.recv(2048)
                if "password".encode() in output or "Password".encode() in output:
                    password = getpass()
                    streams[0].write(f'{password}\n')
                    streams[0].flush()
                    self.logger.log("Installing... this may take some time", color=Colors.yellow)
                    self.logger.debug(streams[1].read().decode())
                    streams = client.exec_command(
                        "su root -c 'apt-get install -f'", get_pty=True)
                    time.sleep(0.2)
                    streams[0].write(f'{password}\n')
                    streams[0].flush()
                    self.logger.debug(streams[1].channel.recv(2048).decode())
                else:
                    self.logger.log("Installing... this may take some time", color=Colors.yellow)
                    self.logger.debug(streams[1].read().decode())
                    self.logger.debug(f"Running command: sudo apt-get install -f")

                    output = client.exec_command(
                        'sudo apt-get install -f')[1]
                    self.logger.debug(output.read().decode())
            return True



import argparse
import os
from pathlib import Path, PurePath
from shutil import copy, copytree, rmtree
import plistlib
import requests
from urllib.parse import urlparse
import zipfile
import subprocess
import tempfile
from subprocess import DEVNULL
from glob import glob

from .ps_copier import Copier
from .ps_downloader import Ldid
from . import __version__
from .ps_utils import Utils
from .ps_logger import Logger, Colors
from .ps_builder import Deb, Control


""" Main Class """


def main(argv=None, in_package=None):
    if argv is None:
        in_package = True

    in_package = False if in_package is None else in_package

    parser = argparse.ArgumentParser()
    parser.add_argument('-d', '--debug', action='store_true',
                        help="shows some debug info, only useful for testing")
    parser.add_argument('-c', '--codesign', action='store_true',
                        help="uses codesign instead of ldid")
    parser.add_argument('-u', '--url', type=str,
                        help="the direct URL of the IPA to be signed")
    parser.add_argument('-p', '--path', type=str,
                        help="the direct local path of the IPA to be signed")
    parser.add_argument('-i', '--install', action='store_true',
                        help="installs the application to your device")
    parser.add_argument('-o', '--output', type=str,
                        help="specify output path")
    parser.add_argument('-b', '--bundleid', type=str,
                        help="specify new bundle id")
    parser.add_argument('-N', '--name', type=str,
                        help="specify new app name")
    parser.add_argument('-m', '--minver', type=str,
                        help="specify new minimum app version (14.0 recommended)")
    parser.add_argument('-v', '--version', action='store_true',
                        help='show current version and exit', )
    parser.add_argument('-l', '--ldidfork', type=str,
                        help="specify a fork of ldid (eg. ProcursusTeam, itsnebulalol [default])")
    parser.add_argument('-f', '--folder', type=str,
                        help="sign multiple IPAs from a direct path to a folder")
    parser.add_argument('-t', '--tcprelay', type=str,
                        help="args for tcprelay rport:lport:host:socketpath (ex: 22:2222:localhost:/var/run/usbmuxd)")
    parser.add_argument('-e', '--entitlements', type=str,
                        help="path to entitlements file")
    parser.add_argument('--no-ldid-check', dest='nocheckldid', action='store_true',
                        help="disable ldid auto update")
    args = parser.parse_args()

    if args.version:
        from permasigner import __version__
        print(f"Permasigner v{__version__.__version__}")
        exit(0)

    ps = Permasigner(in_package, args)
    ps.start()


class Permasigner(object):
    def __init__(self, in_package, args):
        self.in_package = in_package
        self.args = args
        self.utils = Utils(self.args)
        self.logger = Logger(self.args)
        self.outputs = []

    def start(self):
        data_dir = self.utils.get_home_data_directory()
        Path(data_dir).mkdir(exist_ok=True, parents=True)

        if self.in_package:
            self.logger.debug(f"Running from package, not cloned repo.")

        is_extracted = False

        ldid_in_path = self.utils.cmd_in_path('ldid')
        dpkg_in_path = self.utils.cmd_in_path('dpkg-deb')
        git_in_path = self.utils.cmd_in_path('git')

        if git_in_path:
            self.logger.debug(f"Git is in PATH")

            if self.in_package:
                ver_string = f"{__version__.__version__}"
            elif "main" not in subprocess.getoutput([f'git', 'rev-parse', '--abbrev-ref', 'HEAD']):
                ver_string = f"{subprocess.check_output([f'git', 'rev-parse', '--abbrev-ref', 'HEAD']).decode('ascii').strip()}_{subprocess.check_output([f'git', 'rev-parse', '--short', 'HEAD']).decode('ascii').strip()}"
            else:
                ver_string = f"{__version__.__version__}_rev-{subprocess.check_output([f'git', 'rev-parse', '--short', 'HEAD']).decode('ascii').strip()}"
        else:
            self.logger.debug(f"Git is not in PATH")
            if os.environ.get('IS_DOCKER_CONTAINER'):
                ver_string = os.environ.get('VERSION', False)
            else:
                ver_string = f"{__version__.__version__}"

        print(Colors.bold + f"Permasigner | Version {ver_string}")
        print(
            Colors.bold + "Program created by Nebula | Original scripts created by zhuowei | CoreTrust bypass by Linus Henze")
        print()

        # Run checks
        self.checks(ldid_in_path, data_dir)

        # Prompt the user if they'd like to use an external IPA or a local IPA
        if not (self.args.url or self.args.path or self.args.folder):
            option = self.logger.ask("Would you like to use an external or a local IPA? [E, L] ").lower()
            print()

        with tempfile.TemporaryDirectory() as tmpfolder:
            # If the user's choice is external, download an IPA
            # Otherwise, copy the IPA to the temporary directory
            if self.args.url:
                url = self.args.url

                if not PurePath(urlparse(url).path).suffix == ".ipa":
                    self.logger.error("URL provided is not an IPA, make sure to provide a direct link.")
                    exit(1)

                res = requests.get(url, stream=True)
                try:
                    if res.status_code == 200:
                        self.logger.log(f"Downloading file...", color=Colors.yellow)

                        with open(f"{tmpfolder}/app.ipa", "wb") as f:
                            f.write(res.content)
                    else:
                        self.logger.error(f"URL provided is not reachable. Status code: {res.status_code}")
                        exit(1)
                except requests.exceptions.RequestException as err:
                    self.logger.error(f"URL provided is not reachable. Error: {err}")
                    exit(1)
            elif self.args.path:
                path = Path(self.args.path).expanduser()

                if path.exists():
                    path = str(path)
                    if path.endswith(".deb"):
                        self.logger.debug(f"Extracting deb package from {path} to {tmpfolder}/extractedDeb")
                        if dpkg_in_path:
                            self.logger.debug(f"Running command: dpkg-deb -X {path} {tmpfolder}/extractedDeb")
                            subprocess.run(
                                ["dpkg-deb", "-X", path, PurePath(f'{tmpfolder}/extractedDeb')], stdout=DEVNULL)
                        else:
                            deb = Deb(path, PurePath(f'{tmpfolder}/extractedDeb'), self.args)
                            deb.extract()
                        Path(f"{tmpfolder}/app/Payload").mkdir(exist_ok=True)
                        for fname in Path(f"{tmpfolder}/extractedDeb/Applications").iterdir():
                            if fname.name.endswith(".app"):
                                copytree(f"{tmpfolder}/extractedDeb/Applications/{fname}",
                                         f"{tmpfolder}/app/Payload/{fname}")
                                break

                        is_extracted = True
                    elif path.endswith(".ipa"):
                        copy(path, PurePath(f"{tmpfolder}/app.ipa"))
                    else:
                        self.logger.error("That file is not supported by Permasigner! Make sure you're using an IPA or deb.")
                        exit(1)
                else:
                    self.logger.error("That file does not exist! Make sure you're using a direct path to the IPA file.")
                    exit(1)
            elif self.args.folder:
                for fpath in glob(f"{self.args.folder}/*.ipa"):
                    if Path(f"{tmpfolder}/app.ipa").exists():
                        Path(f"{tmpfolder}/app.ipa").unlink()
                    if Path(f"{tmpfolder}/app").exists():
                        rmtree(f"{tmpfolder}/app")
                    if Path(f"{tmpfolder}/deb").exists():
                        rmtree(f"{tmpfolder}/deb")

                    fname = PurePath(fpath).name
                    self.logger.log(f"Signing {fname}...", color=Colors.yellow)
                    print()

                    copy(fpath, f"{tmpfolder}/app.ipa")
                    out_dir = self.run(tmpfolder, ldid_in_path, dpkg_in_path, data_dir, is_extracted)
                    self.outputs.append(out_dir)
            elif option == "e":
                url = self.logger.ask("Paste in the *direct* path to an IPA online: ")

                if not PurePath(urlparse(url).path).suffix == ".ipa":
                    self.logger.error("URL provided is not an IPA, make sure to provide a direct link.")
                    exit(1)
                print()

                res = requests.get(url, stream=True)
                try:
                    if res.status_code == 200:
                        self.logger.log(f"Downloading file...", color=Colors.yellow)

                        with open(f"{tmpfolder}/app.ipa", "wb") as f:
                            f.write(res.content)
                    else:
                        self.logger.error(f"URL provided is not reachable. Status code: {res.status_code}")
                        exit(1)
                except requests.exceptions.RequestException as err:
                    self.logger.error(f"URL provided is not reachable. Error: {err}")
                    exit(1)
            elif option == "l":
                if os.environ.get('IS_DOCKER_CONTAINER', False):
                    self.logger.log(
                        "Running in Docker container, please place an IPA in the 'ipas' folder, then put the name of the file below.",
                        color=Colors.orange)
                    ipa_name = self.logger.ask('    IPA name (ex. Taurine.ipa, DemoApp.ipa): ')
                    path = f"/permasigner/ipas/{ipa_name}"
                else:
                    path = self.logger.ask("Paste in the path to an IPA in your file system: ").replace(' ', '').strip("'")

                if path.startswith('~'):
                    path = Path(path).expanduser()

                if Path(path).exists():
                    if PurePath(path).suffix == ".deb":
                        if dpkg_in_path:
                            self.logger.debug(f"Running command: dpkg-deb -X {path} {tmpfolder}/extractedDeb")
                            subprocess.run(
                                ["dpkg-deb", "-X", path, f"{tmpfolder}/extractedDeb"], stdout=DEVNULL)
                        else:
                            deb = Deb(path, f"{tmpfolder}/extractedDeb", self.args)
                            deb.extract()
                        self.logger.debug(f"Extracted deb file from {path} to {tmpfolder}/extractedDeb")
                        Path(f"{tmpfolder}/app/Payload").mkdir(exist_ok=False)
                        for fname in Path(f"{tmpfolder}/extractedDeb/Applications").iterdir():
                            if fname.name.endswith(".app"):
                                copytree(f"{tmpfolder}/extractedDeb/Applications/{fname}",
                                         f"{tmpfolder}/app/Payload/{fname}")

                        is_extracted = True
                    elif PurePath(path).suffix == ".ipa":
                        copy(path, f"{tmpfolder}/app.ipa")
                    else:
                        self.logger.error("That file is not supported by Permasigner! Make sure you're using an IPA or deb.")
                        exit(1)
                else:
                    self.logger.error("That file does not exist! Make sure you're using a direct path to the IPA file.")
                    exit(1)
            else:
                self.logger.error(f"That is not a valid option!")
                exit(1)

            if not self.args.folder:
                print()

            is_installed = False
            if not self.args.folder:
                out_dir = self.run(tmpfolder, ldid_in_path, dpkg_in_path, data_dir, is_extracted)

                if self.args.install:
                    is_installed = self.install(out_dir)

            # Done, print end message
            self.logger.log(f"We are finished!", color=Colors.green)

            if is_installed:
                self.logger.log(f"The application was installed to your device, no further steps are required!",
                                color=Colors.green)
            else:
                self.logger.log(
                    f"Copy the newly created deb from the output folder to your jailbroken iDevice and install it!",
                    color=Colors.green)

            self.logger.log(f"The app will continue to work when rebooted to stock.", color=Colors.green)
            self.logger.log(
                f"Also, this is free and open source software! Feel free to donate to my Patreon if you enjoy :)",
                color=Colors.green)
            print(Colors.green + "    https://patreon.com/nebulalol" + Colors.reset)
            if self.args.folder:
                final_outputs = ""
                for output in self.outputs:
                    if final_outputs == "":
                        final_outputs = f"{output}"
                    else:
                        final_outputs = f"{final_outputs}, {output}"
                self.logger.log(f"Output files: {final_outputs}", color=Colors.green)
            else:
                self.logger.log(f"Output file: {out_dir}", color=Colors.green)

    def checks(self, ldid_in_path, data_dir):
        # Check if codesign arg is added on Linux or iOS
        if self.args.codesign:
            if not self.utils.is_macos():
                self.logger.error(f"Codesign can only be used on macOS, remove the argument to use ldid instead.")
                exit(1)

        # Auto download ldid
        if not ldid_in_path and not self.args.nocheckldid:
            name = 'ldid'
            if self.utils.is_windows():
                name = 'ldid.exe'
            if Path(data_dir).joinpath(name).exists():
                ldid = Ldid(data_dir, name, self.args, self.utils, True)
            else:
                self.logger.log("ldid binary is not found, downloading latest binary.", color=Colors.yellow)
                ldid = Ldid(data_dir, name, self.args, self.utils, False)
            ldid.download()

    def install(self, out_dir):
        if not self.utils.is_ios():
            from .ps_installer import Installer

        if not self.utils.is_ios():
            installer = Installer(self.args, out_dir)
            is_installed = installer.install_deb()
        else:
            print("Checking if user is in sudoers")
            p = subprocess.run('sudo -nv'.split(),
                               capture_output=True)
            if p.returncode == 0 or 'password' in p.stderr.decode():
                print("User is in sudoers, using sudo command")
                self.logger.debug(f"Running command: sudo dpkg -i {out_dir}")

                subprocess.run(
                    ["sudo", "dpkg", "-i", f"{out_dir}"], capture_output=True)

                subprocess.run(
                    ['sudo', 'apt-get', 'install', '-f'], capture_output=True)

                is_installed = True
            else:
                print("User is not in sudoers, using su instead")
                self.logger.debug(f"Running command: su root -c 'dpkg -i {out_dir}")

                subprocess.run(
                    f"su root -c 'dpkg -i {out_dir}'".split(), capture_output=True)

                subprocess.run(
                    "su root -c 'apt-get install -f'".split(), capture_output=True)

                is_installed = True

        return is_installed

    def run(self, tmpfolder, ldid_in_path, dpkg_in_path, data_dir, is_extracted):
        # Unzip the IPA file
        if not is_extracted:
            self.logger.log(f"Unzipping IPA...", color=Colors.yellow)
            with zipfile.ZipFile(PurePath(f'{tmpfolder}/app.ipa'), 'r') as f:
                with Path(f"{tmpfolder}/app") as path:
                    path.mkdir(exist_ok=False)
                    f.extractall(path)
                    for ds in Path(path).rglob('.DS_Store*'):
                        Path(ds).unlink()
            print()

        # Read data from the plist
        app_dir = ''
        payload = Path(tmpfolder).joinpath('app').joinpath('Payload')
        if payload.exists():
            for fname in payload.rglob('*.app'):
                app_dir = fname
                break

            if app_dir == '':
                self.logger.error("Unable to find the application bundle")
                exit(1)
            print("Found app directory!")
        else:
            self.logger.error(f"IPA/deb is not valid!")
            exit(1)

        if Path(app_dir).joinpath('Info.plist').exists():
            self.logger.log(f"Reading plist...", color=Colors.yellow)
            with open(f'{app_dir}/Info.plist', 'rb') as f:
                info = plistlib.load(f)
                if self.args.name:
                    app_name = self.args.name
                else:
                    app_name = info["CFBundleName"]

                if self.args.bundleid:
                    app_bundle = self.args.bundleid
                elif "science.xnu.undecimus" in info["CFBundleIdentifier"]:
                    app_bundle = "8H69ZMY835.science.xnu.undecimus.8H69ZMY835"
                else:
                    app_bundle = info["CFBundleIdentifier"]

                app_version = info["CFBundleShortVersionString"]

                if self.args.minver:
                    app_min_ios = self.args.minver
                else:
                    app_min_ios = info["MinimumOSVersion"]

                app_author = app_bundle.split(".")[1]

                app_executable = info["CFBundleExecutable"]
                print("Found information about the app!")
        else:
            self.logger.error("Unable to find Info.plist, can't read application data")
            exit(1)

        print()

        # Get the deb file ready
        self.logger.log(f"Preparing deb file...", color=Colors.yellow)
        print("Making directories...")
        Path(tmpfolder).joinpath('deb').joinpath('Applications').mkdir(exist_ok=False, parents=True)
        Path(tmpfolder).joinpath('deb').joinpath('DEBIAN').mkdir(exist_ok=False, parents=True)
        print("Copying deb file scripts and control...")
        copier = Copier(app_name, app_bundle, app_version, app_min_ios, app_author, self.in_package)
        postrm_path = PurePath(f'{tmpfolder}/deb/DEBIAN/postrm')
        postinsts_path = PurePath(f'{tmpfolder}/deb/DEBIAN/postinst')
        control_path = PurePath(f'{tmpfolder}/deb/DEBIAN/control')
        copier.copy_postinst(postinsts_path)
        copier.copy_control(control_path)
        copier.copy_postrm(postrm_path)
        print("Changing deb file scripts permissions...")
        self.utils.set_executable_permission(postrm_path)
        self.utils.set_executable_permission(postinsts_path)
        print("Copying app files...")
        full_app_path = PurePath(f"{tmpfolder}/deb/Applications/{app_dir.name}")
        copytree(app_dir, full_app_path)
        print("Changing app executable permissions...")
        self.utils.set_executable_permission(PurePath(f'{full_app_path}/{app_executable}'))
        print()

        # Sign the app
        self.logger.log(f"Signing app...", color=Colors.yellow)
        copier.copy_entitlements(f"{tmpfolder}/entitlements.plist")
        if self.in_package:
            cert_path = self.utils.get_resource_path(__name__, "data/certificate.p12")
        else:
            cert_path = PurePath(f'{Path.cwd()}/permasigner/data/certificate.p12')

        if self.args.codesign:
            print("Signing with codesign as it was specified...")
            subprocess.run(
                ['security', 'import', cert_path, '-P', 'password', '-A'], stdout=DEVNULL)

            subprocess.run(['codesign', '-s', 'We Do A Little Trolling iPhone OS Application Signing',
                            '--force', '--deep', '--preserve-metadata=entitlements', f'{full_app_path}'],
                           stdout=DEVNULL)
        else:
            print("Signing with ldid...")
            if self.utils.is_ios() or ldid_in_path:
                ldid_cmd = 'ldid'
            elif self.utils.is_windows():
                ldid_cmd = PurePath(f'{data_dir}/ldid.exe')
            else:
                ldid_cmd = PurePath(f'{data_dir}/ldid')

            self.logger.debug(
                f"Running command: {ldid_cmd} -S{PurePath(f'{tmpfolder}/entitlements.plist')} -M -K{cert_path} -Upassword '{full_app_path}'")

            subprocess.run([f'{ldid_cmd}', f'-S{tmpfolder}/entitlements.plist', '-M',
                            f'-K{cert_path}', '-Upassword', f'{full_app_path}'], stdout=DEVNULL)

            if self.args.entitlements:
                self.logger.debug(f"Signing with extra entitlements located in {self.args.entitlements}")
                subprocess.run([f'{ldid_cmd}', f'-S{self.args.entitlements}', '-M',
                                f'-K{cert_path}', '-Upassword', f'{full_app_path}'], stdout=DEVNULL)
        print()

        # Package the deb file
        self.logger.log(f"Packaging the deb file...", color=Colors.yellow)
        if self.args.output:
            out_dir = self.args.output
            Path(out_dir).mkdir(exist_ok=True)
        elif self.in_package:
            Path(f"{data_dir}/output").mkdir(exist_ok=True)
            out_dir = PurePath(f"{data_dir}/output")
        else:
            Path("output").mkdir(exist_ok=True)
            out_dir = PurePath(f"{Path.cwd()}/output")

        if dpkg_in_path:
            out_path = PurePath(f"{out_dir}/{app_name.replace(' ', '') + f'_{app_version}' + '.deb'}")
            dpkg_cmd = f"dpkg-deb -Zxz --root-owner-group -b {tmpfolder}/deb {out_path}"
            self.logger.debug(f"Running command: {dpkg_cmd}")
            subprocess.run(dpkg_cmd.split(), stdout=DEVNULL)
            return out_path
        else:
            control = Control(app_bundle, app_version, app_min_ios, app_name, app_author, app_executable)
            deb = Deb(f"{tmpfolder}/deb/Applications/", out_dir, self.args)
            output_name = deb.build(PurePath(f"{tmpfolder}/deb/DEBIAN/postinst"),
                                    PurePath(f"{tmpfolder}/deb/DEBIAN/postrm"), control)
            return PurePath(f'{out_dir}/{output_name}')


if __name__ == "__main__":
    main()



from pathlib import Path, PurePath
import requests
import hashlib
import platform
from shutil import move
from requests.exceptions import RequestException, ConnectionError
from urllib3.exceptions import NewConnectionError

from .ps_logger import Logger, Colors


class Hash(object):
    def __init__(self, args):
        self.args = args
        self.logger = Logger(self.args)

    def get_hash(self, filePath, url):
        m = hashlib.md5()
        if url is None:
            with open(filePath, 'rb') as fh:
                m = hashlib.md5()
                while True:
                    data = fh.read(8192)
                    if not data:
                        break
                    m.update(data)
                return m.hexdigest()
        else:
            try:
                res = requests.get(url, stream=True)
                if res.status_code == 200:
                    content = bytearray()
                    for data in res.iter_content(4096):
                        content += data
                        m.update(data)
                    return m.hexdigest(), content
                else:
                    return m.hexdigest(), None
            except (NewConnectionError, ConnectionError, RequestException) as err:
                self.logger.error(f"ldid download URL is not reachable. Error: {err}")
                return m.hexdigest(), None


class Ldid(object):
    def __init__(self, data_dir, name, args, utils, exists=False):
        self.name = name
        self.data_dir = data_dir
        self.args = args
        self.utils = utils
        self.ldid_fork = "itsnebulalol"
        self.exists = exists
        self.logger = Logger(self.args)
        self.hash = Hash(self.args)

    def get_arch(self):
        if self.utils.is_linux() and platform.machine() == "x86_64":
            return "ldid_linux_x86_64"
        elif self.utils.is_linux() and platform.machine() == "aarch64":
            return "ldid_linux_aarch64"
        elif self.utils.is_linux() and platform.machine() == "armv7l":
            return "ldid_linux_armv7l"
        elif self.utils.is_macos() and platform.machine() == "x86_64":
            return "ldid_macos_x86_64"
        elif self.utils.is_macos() and platform.machine() == "arm64":
            return "ldid_macos_arm64"
        elif self.utils.is_windows() and platform.machine() in ["AMD64", "x86_64"]:
            return "ldid_win32_x86_64.exe"
        elif self.utils.is_freebsd12() and platform.machine() == "amd64":
            return "ldid_freebsd12_amd64"
        elif self.utils.is_freebsd13 and platform.machine() == "amd64":
            return "ldid_freebsd13_amd64"

    def process(self, content):
        self.logger.log(f"ldid is outdated or malformed, downloading latest version...", color=Colors.yellow)

        if content is not None:
            with open(self.name, "wb") as f:
                f.write(content)
                self.logger.debug(f"Wrote file.")

            if self.exists:
                self.logger.debug("Removing outdated version of ldid")
                Path(self.data_dir).joinpath(self.name).unlink()

            self.utils.set_executable_permission(self.name)
            destination = PurePath(self.data_dir).joinpath(self.name)
            move(self.name, destination)
            self.logger.debug(f"Moved ldid to {str(destination)}")

    def download(self):
        if self.args.ldidfork:
            ldid_fork = self.args.ldidfork
        else:
            ldid_fork = self.ldid_fork

        self.logger.debug(f"Using ldid fork {ldid_fork}.")

        url = f"https://github.com/{ldid_fork}/ldid/releases/latest/download/{self.get_arch()}"

        self.logger.debug(f"Comparing {self.get_arch()} hash with {url}")

        remote_hash, content = self.hash.get_hash(None, url)
        local_hash = None

        if self.exists:
            local_hash = self.hash.get_hash(PurePath(self.data_dir).joinpath(self.name), None)

        if remote_hash == local_hash:
            self.logger.debug(f"ldid hash successfully verified.")
        else:
            if content is None:
                if self.exists:
                    self.logger.log('Could not verify remote hash, falling back to ldid found in path',
                                    color=Colors.yellow)
                else:
                    self.logger.error('Download url is not reachable, and no ldid found in path, exiting.')
                    exit(1)
            else:
                self.logger.debug(f"Ldid hash failed to verify, saving newer version")
                self.process(content)



import os
from setuptools import setup, find_packages

here = os.path.abspath(os.path.dirname(__file__))

with open(os.path.join(here, "README.md"), encoding="utf-8") as f:
    readme = f.read()

with open(os.path.join(here, "permasigner", "__version__.py"), encoding="utf-8") as f:
    version = f.read()

version = version.split('=')[-1].strip().strip('"').strip("'")


with open(os.path.join(here, "requirements.txt"), encoding="utf-8") as f:
    requires = f.readlines()


setup(
    name='permasigner',
    version=version,
    description=('Permanently signs IPAs on jailbroken iDevices (persists on stock).'),
    license='BSD-3-Clause',
    url='https://github.com/itsnebulalol/permasigner',
    python_requires=">=3.7",
    packages=find_packages(),
    long_description=readme,
    long_description_content_type="text/markdown",
    classifiers=[
        'Natural Language :: English',
        'Environment :: Console',
        'Operating System :: OS Independent',
    ],
    entry_points={
        'console_scripts': [
            'permasigner = permasigner.__main__:main',
        ],
    },
    keywords='python, windows, macos, linux, docker, cli, open-source, ios, command-line-app, cli-app, hacktoberfest, procursus, permasign, permasigner',
    include_package_data=True,
    author='Nebula',
    install_requires=requires,
    author_email='me@itsnebula.net',
    zip_safe=True,
)



import math

FIELD_BUILT_USING = "Built-Using"
FIELD_HOMEPAGE = "Homepage"
FIELD_DESCRIPTION = "Description"
FIELD_MAINTAINER = "Maintainer"
FIELD_INSTALLED_SIZE = "Installed-Size"
FIELD_PROVIDES = "Provides"
FIELD_CONFLICTS = "Conflicts"
FIELD_BREAKS = "Breaks"
FIELD_PRE_DEPENDS = "Pre-Depends"
FIELD_REPLACES = "Replaces"
FIELD_ENHANCES = "Enhances"
FIELD_SUGGESTS = "Suggests"
FIELD_RECOMMENDS = "Recommends"
FIELD_DEPENDS = "Depends"
FIELD_ESSENTIAL = "Essential"
FIELD_ARCHITECTURE = "Architecture"
FIELD_PRIORITY = "Priority"
FIELD_SECTION = "Section"
FIELD_VERSION = "Version"
FIELD_SOURCE = "Source"
FIELD_PACKAGE = "Package"
FIELD_NAME = "Name"
FIELD_AUTHOR = "Author"
FIELD_TAGS = "Tags"
FIELD_DEPICTION = "Depiction"

BINARY_CONTROL_ALLOWED_FIELDS = (
    FIELD_DEPICTION,
    FIELD_TAGS,
    FIELD_AUTHOR,
    FIELD_NAME,
    FIELD_PACKAGE,
    FIELD_SOURCE,
    FIELD_VERSION,
    FIELD_SECTION,
    FIELD_PRIORITY,
    FIELD_ARCHITECTURE,
    FIELD_ESSENTIAL,
    FIELD_DEPENDS,
    FIELD_RECOMMENDS,
    FIELD_SUGGESTS,
    FIELD_ENHANCES,
    FIELD_REPLACES,
    FIELD_PRE_DEPENDS,
    FIELD_BREAKS,
    FIELD_CONFLICTS,
    FIELD_PROVIDES,
    FIELD_INSTALLED_SIZE,
    FIELD_MAINTAINER,
    FIELD_DESCRIPTION,
    FIELD_HOMEPAGE,
    FIELD_BUILT_USING
)

BINARY_CONTROL_REQUIRED_FIELDS = (FIELD_PACKAGE, FIELD_VERSION, FIELD_ARCHITECTURE, FIELD_MAINTAINER, FIELD_DESCRIPTION)

BINARY_DEPENDS_ET_AL_FIELDS = (FIELD_DEPENDS, FIELD_RECOMMENDS, FIELD_SUGGESTS, FIELD_ENHANCES, FIELD_PRE_DEPENDS,
                               FIELD_BREAKS, FIELD_CONFLICTS, FIELD_PROVIDES, FIELD_REPLACES)


class BinaryControl(object):
    def __init__(self, package, version, architecture, maintainer, description):
        self._control_fields = {
            FIELD_PACKAGE: package,
            FIELD_VERSION: version,
            FIELD_ARCHITECTURE: architecture,
            FIELD_MAINTAINER: maintainer,
            FIELD_DESCRIPTION: description
        }
        self.installed_size_bytes = None

    def get_control_line(self, field_name):
        if field_name not in self._control_fields:
            raise KeyError("{} is not set.".format(field_name))

        field_value = self._control_fields[field_name]

        return "{}: {}".format(field_name, field_value)

    def get_control_line_value(self, field_name):
        if field_name not in self._control_fields:
            raise KeyError("{} is not set.".format(field_name))

        return self._control_fields[field_name]

    def set_control_field(self, field_name, field_value):
        if field_name not in BINARY_CONTROL_ALLOWED_FIELDS:
            raise KeyError("{} is not a valid field for a binary control file".format(field_name))

        self._control_fields[field_name] = field_value

    def set_control_fields(self, new_control_fields):
        for field_name, field_value in new_control_fields.items():
            self.set_control_field(field_name, field_value)

    def get_calculated_install_size_line(self):
        if self.installed_size_bytes is None:
            return None

        return '{}: {}'.format(FIELD_INSTALLED_SIZE, int(math.ceil(self.installed_size_bytes / 1024.0)))

    def get_control_text(self):
        control_lines = []
        for field_name in BINARY_CONTROL_REQUIRED_FIELDS:
            control_lines.append(self.get_control_line(field_name))

        if FIELD_INSTALLED_SIZE not in self._control_fields:
            install_size_line = self.get_calculated_install_size_line()

            if install_size_line is not None:
                control_lines.append(install_size_line)

        for field_name in BINARY_CONTROL_ALLOWED_FIELDS:
            if field_name in BINARY_CONTROL_REQUIRED_FIELDS:
                continue

            if field_name not in self._control_fields:
                continue

            control_lines.append(self.get_control_line(field_name))

        return "\n".join(control_lines) + "\n"

    def is_field_defined(self, field_name):
        return field_name in self._control_fields

    def get_default_output_name(self):
        package = self._control_fields[FIELD_PACKAGE]
        version = self._control_fields[FIELD_VERSION]
        architecture = self._control_fields[FIELD_ARCHITECTURE]

        return '{}_{}_{}.deb'.format(package, version, architecture)



import sys
import platform
import os
import subprocess
import importlib
from importlib import util
from pathlib import Path, PurePath
from shutil import which

from permasigner.ps_logger import Logger


class Utils(object):
    def __init__(self, args):
        self.args = args
        self.logger = Logger(self.args)

    @staticmethod
    def is_ios():
        if not sys.platform == "darwin":
            return False

        return platform.machine().startswith("i")

    @staticmethod
    def is_macos():
        if platform.machine().startswith("i"):
            return False

        return sys.platform == "darwin"

    @staticmethod
    def is_linux():
        return sys.platform == "linux"

    @staticmethod
    def is_windows():
        return sys.platform == "win32"

    @staticmethod
    def is_freebsd12():
        return sys.platform == "freebsd12"

    @staticmethod
    def is_freebsd13():
        return sys.platform == "freebsd13"

    @staticmethod
    def is_dpkg_installed(pkg):
        return (os.system("dpkg -s " + pkg + "> /dev/null 2>&1")) == 0

    def set_executable_permission(self, path):
        file = Path(path)
        mode = file.stat().st_mode
        mode |= (mode & 0o444) >> 2
        file.chmod(mode)
        self.logger.debug(f'Set chmod +x on {file}')

    def cmd_in_path(self, cmd):
        self.logger.debug(f"Checking if command {cmd} is in PATH...")

        path = which(cmd)

        if cmd == "ldid":
            if self.is_ios():
                self.logger.debug(f"Checking for ldid on iOS")

                if Path("/.bootstrapped").exists():
                    self.logger.error(
                        "Your device seems to be strapped with Elucubratus. Unfortunately, we do not support these devices. You can switch to a device that uses Procursus (Taurine, odysseyra1n), or use the online method on our GitHub.")
                    print("    https://github.com/itsnebulalol/permasigner/wiki/Run-Online")
                    exit(1)

                if self.is_dpkg_installed("ldid"):
                    self.logger.debug(f"ldid is installed via dpkg")
                    return True
                else:
                    self.logger.error("ldid is required on iOS, but it is not installed. Please install it from Procursus.")
                    exit(1)

            if path is not None:
                if "procursus" not in subprocess.getoutput(path):
                    return False

                self.logger.debug(f"Procursus ldid is installed")
                return True

            return False

        return path is not None

    def get_home_data_directory(self):
        ps_home = os.environ.get("PERMASIGNER_HOME")
        if ps_home:
            return ps_home

        user_home = Path.home()
        if self.is_linux() or self.is_freebsd12() or self.is_freebsd13():
            xdg_data_home = os.environ.get("XDG_DATA_HOME", PurePath(f'{user_home}/.local/share'))
            return PurePath(f'{xdg_data_home}/permasigner')
        elif self.is_ios() or self.is_macos():
            return PurePath(f'{user_home}/Library/Application Support/permasigner')
        elif self.is_windows():
            return PurePath(f"{os.getenv('APPDATA')}/permasigner")

    @staticmethod
    def get_resource_path(package, resource):
        spec = importlib.util.find_spec(package)
        if spec is None:
            return None
        loader = spec.loader
        if loader is None or not hasattr(loader, 'get_data'):
            return None
        mod = (sys.modules.get(package) or
               importlib._bootstrap._load(spec))
        if mod is None or not hasattr(mod, '__file__'):
            return None

        parts = resource.split('/')
        parts.insert(0, PurePath(mod.__file__).parent)
        return f"{PurePath(*parts)}"
